# Pagean

Pagean is a web page analysis tool designed to automate tests requiring web pages to be loaded in a browser window (e.g. 404 error loading an external resource, page renders with horizontal scrollbars).  The specific tests are outlined below, but are all general tests that do not include any page-specific logic.

**Note: As of v4.0.0, this module is being being released as `pagean` and `page-load-tests` has been deprecated. This drove breaking configuration changes that will require updating your project's configuration file. See the [version upgrade guide](https://gitlab.com/gitlab-ci-utils/pagean/-/blob/master/docs/upgrade-guide.md) for complete details.**

## Installation

Install Pagean globally (as shown below), or locally, via [npm](https://www.npmjs.com/).

```
npm install -g pagean
```

## Usage

Pagean runs as a command line tool and is executed as follows:

```
Installed globally:
> pagean [options]

Installed locally:
> npx pagean [options]

Options:
  -V, --version        output the version number
  -c, --config <file>  the path to the pagean configuration file (default: "./.pageanrc.json")
  -h, --help           display help for command
```

Pagean requires a configuration file named, which can be specified via the CLI as detailed above, or use the default file `.pageanrc.json` in the project root.  This file provides the URLs to be tested and options to configure the tests and reports.  Details on the available tests and the configuration file format are provided below.

## Test Cases

The tests use [Puppeteer](https://github.com/GoogleChrome/puppeteer) to launch a headless Chrome browser.  The URLs defined in the configuration file are each loaded once, and after page load the applicable tests are executed.  Test results are `passed` or `failed`, but can be configured to report `warning` instead of failure.  Only a `failed` test will cause the test process to fail and exit with an error code (a `warning` will not).

### Horizontal Scrollbar Test

The horizontal scrollbar test fails if the rendered page has a horizontal scrollbar.  If a specific browser viewport size is desired for this test, that can be configured in the `puppeteerLaunchOptions`.

### Console Output Test

The console output test fails if any output is written to the browser console.  An array is included in the report with all entries, as shown below:

```json
Received array:
[
    {
        "_args": [],
        "_location": {
            "lineNumber": undefined,
            "url": "https://this.url.does.not.exist/file.js"
        },
        "_text": "Failed to load resource: net::ERR_NAME_NOT_RESOLVED",
        "_type": "error"
    }
]
```

### Console Error Test

The console error test fails if any error is written to the browser console, but is otherwise the same as the console output test.  This separation allows for testing for console errors, but allowing any other console output.

### Rendered HTML Test

The rendered HTML test is intended for cases where content is dynamically created prior to page load (i.e. the `load` event firing).  The rendered HTML is returned and checked with [HTML Hint](https://www.npmjs.com/package/htmlhint) and the test fails if any issues are found.  An array is included in the report with all HTML Hint issues, as shown below:

```json
Received array:
[
    {
        "col": 9,
        "evidence": "    <div id=\"div1\"></div>",
        "line": 6,
        "message": "The id value [ div1 ] must be unique.",
        "raw": " id=\"div1\"",
        "rule": {
            "description": "The value of id attributes must be unique.",
            "id": "id-unique",
            "link": "https://github.com/thedaviddias/HTMLHint/wiki/id-unique"
        },
        "type": "error"
    }
]
```

An htmlhintrc file can be specified in the configuration file, otherwise the default "./.htmlhintrc" file will be used (if it exists). See the Configuration section below.

Note: This test may not find some errors in the original HTML that are removed/resolved as the page is parsed (e.g. closing tags with no opening tags).

### Page Load Time Test

The page load time test fails if the page load time (from start through the `load` event) exceeds the defined threshold in the configuration file (or the default of 2 seconds).  The actual load time is included in the report.  Tests will time out at twice the page load time threshold.

### External Script Test

The external script test is intended to identify any externally loaded javascript files (e.g. loaded from a CDN) and aggregate those files so they can undergo further analysis (e.g. dependency vulnerability scanning). The test is included here since these tests load fully rendered pages, therefore allowing the aggregation of this data for pages generated using any language or framework.  By default the test returns a warning if the page includes any javascript files loaded from a different domain than the page (although this could be overridden to fail instead via setting `failWarn: false`, see the Configuration section below).  These files are then downloaded and saved in the "pagean-external-files" directory in the project root.  Subdirectories are created for each domain, then following the URL path.  For example, the following script...

```html
<script src="https://bootstrapcdn.com/bootstrap/4.5.0/js/bootstrap.min.js"></script>
```

...will be saved as `./bootstrapcdn.com/bootstrap/4.5.0/js/bootstrap.min.js`.  The `data` array in the test report includes the original files URL and the local saved filename.  Each external script is saved only once, but will be reported on any page where it is referenced.

## Reports

Based on the `reporters` configuration, Pagean results may be displayed in the console and saved in two reports in the project root directory (any or all of the three):

- A JSON report named [`pagean-results.json`](https://gitlab-ci-utils.gitlab.io/pagean/pagean-results.json)
- An HTML report named [`pagean-results.html`](https://gitlab-ci-utils.gitlab.io/pagean/pagean-results.html)

Both reports contain:

- The time of test execution
- A summary of the total tests and results (passed, warning, failed)
- The detailed test results, including the URL tested, list of tests performed on that URL with results, and, if applicable, any relevant data associated with the test failure (e.g. the console errors if the console error test fails).

Complete reports for the example case in this project (the tests as specified in the project [`.pageanrc.json`](https://gitlab.com/gitlab-ci-utils/pagean/-/blob/master/.pageanrc.json) file) can be found at the links above.

## Configuration

Pagean looks for a configuration file as specified via the CLI, or defaults to a file named `.pageanrc.json` in the project root.  If the configuration file is not found, is not valid JSON, or does not contain any URLs to check the job will fail.

Below is an example `.pageanrc.json` file, which is broken into six major properties:

- `htmlhintrc`: An optional path to an htmlhintrc file to be used in the rendered HTML test
- `project`: An optional name of the project, which is included in HTML and JSON reports.
- `puppeteerLaunchOptions`: An optional set of options to pass to Puppeteer on launch.  There are no default options. The complete list of available options can be found at https://github.com/GoogleChrome/puppeteer/blob/master/docs/api.md#puppeteerlaunchoptions.
- `reporters`:  An optional array of reporters indicating the test reports that should be provided.  There are three possible options - `cli`, `html`, and `json`. The `cli` option reports all test details to the console, but the final results summary is always output (even with `cli` disabled).  If `reporters` is specified, at least one reporter must be included. The default value, as specified below, is all three reporters enabled.
- `settings`: These settings enable/disable or configure tests, and are applied to all tests overriding the default values.
  - The shorthand notation allows easy enabling/disabling of tests. In this format the test name is given with a boolean value to enable or disable the test.  In this case any other test-specific settings use the default values.
  - The longhand version includes an object for each test. Every test includes two possible properties (some tests include additional settings):
    - `enabled`: A boolean value to enable/disable the test, and some tests include additional settings (default `true` for all tests).
    - `failWarn`: A boolean value causing a failed test to report a warning instead of failure.  A warning result will not cause the test process to fail (exit with an error code).  The default value for all tests is `false` except the `externalScriptTest`, as shown below.

The shorthand:

```json
"settings": {
    "consoleErrorTest": true
}
```

is equivalent to the longhand:

```json
"settings": {
    "consoleErrorTest": {
        "enabled": true,
        "failWarn": false
    }
}
```

All available settings with the default values are shown below.

- `urls`: An array of URLs to be tested, which must contain at least one value.  Each array entry can either be a URL string, or an object that contains a `url` string and an optional `settings` object.  This object can contain any of the `settings` values identified above and will override that setting for testing that URL.  The `url` string can be either an actual URL or a local file, as shown in the example below.

```json
{
    "puppeteerLaunchOptions": {
        "args": [ "--no-sandbox" ]
    },
    "reporters": [
        "cli",
        "html",
        "json"
    ],
    "settings": {
        "horizontalScrollbarTest": {
            "enabled": true,
            "failWarn": false
        },
        "consoleOutputTest":  {
            "enabled": true,
            "failWarn": false
        },
        "consoleErrorTest":  {
            "enabled": true,
            "failWarn": false
        },
        "renderedHtmlTest":  {
            "enabled": true,
            "failWarn": false
        },
        "pageLoadTimeTest":  {
            "enabled": true,
            "failWarn": false,
            "pageLoadTimeThreshold": 2
        },
        "externalScriptTest":  {
            "enabled": true,
            "failWarn": true
        }
    },
    "urls": [
        "https://gitlab.com/gitlab-ci-utils/pagean/",
        {
            "url": "./tests/test-cases/consoleLog.html",
            "settings": {
                "consoleOutputTest": false
            }
        }
    ]
}
```

## Docker Images

Provided with the Pagean project are Docker images configured to run the tests.  All available image tags can be found in the `gitlab-ci-utils/pagean` repository at https://gitlab.com/gitlab-ci-utils/pagean/container_registry.  Details on each release can be found on the [Releases](https://gitlab.com/gitlab-ci-utils/pagean/releases) page.

**Note:** Any images in the `gitlab-ci-utils/pagean/tmp` repository are temporary images used during the build process and may be deleted at any point.

## GitLab CI Configuration

The following is an example job from a .gitlab-ci.yml file to use this image to run Pagean against another project in GitLab CI:

```yaml
pagean:
  image: registry.gitlab.com/gitlab-ci-utils/pagean:latest
  stage: test
  script:
    - pagean
  artifacts:
    when: always
    paths:
      - pagean-results.html
      - pagean-results.json
      - pagean-external-scripts/
```

## Linting Pageanrc Files

A command line tool is also available to lint pageanrc files, which is executed as follows:

```
Installed globally:
> pageanrc-lint [options] [file] (default: "./.pageanrc.json")

Installed locally:
> npx pageanrc-lint [options] [file] (default: "./.pageanrc.json")

Lint a pageanrc file

Options:
  -V, --version  output the version number
  -j, --json     output JSON with full details
  -h, --help     display help for command
```

The `--json` option outputs the JSON results to stdout in all cases for consistency (`[]` if no errors found). Otherwise errors are output to stderr, for example:

```sh
.\tests\test-configs\cli-tests\all-errors.pageanrc.json
  <pageanrc>.puppeteerLaunchOptions                  should NOT have fewer than 1 items
  <pageanrc>.reporters[0]                            should be equal to one of the allowed values (cli, html, json)
  <pageanrc>.settings.consoleOutputTest              should be either boolean or object with the appropriate properties
  <pageanrc>.settings.pageLoadTimeTest.some          should NOT contain additional properties: "some"
  <pageanrc>.urls[2].settings.consoleOutputTest      should be either boolean or object with the appropriate properties
  <pageanrc>.urls[3]                                 should be either URL string or object with the appropriate properties
  <pageanrc>.urls[5]                                 should have required property url
```

Note that because of the large number of options, which are dependent on an external project, the linting of `puppeteerLaunchOptions` only checks that at least one property is provided, it does not check the detailed settings.
